"""
functions that allow you to generate CLI parameters from a confu schema for
argparse or click
"""

try:
    import click
except ImportError:
    pass


def option_name(path, delimiter="--"):
    """
    Returns a cli option name from attribute path

    **Arguments**

    - path (`list`): attribute path
    - delimiter (`str`): delimiter for nested attributes

    **Returns**

    cli option name (`str`)
    """

    return "--{}".format(delimiter.join(path).replace("_", "-"))


def destination_name(path, delimiter="__"):
    """
    Returns a cli option destination name from attribute path

    **Arguments**

    - path (`list`): attribute path
    - delimiter (`str`): delimiter for nested attributes

    **Returns**

    cli destination name (`str`)
    """
    return "{}".format(delimiter.join(path))


def default(value, path, defaults):
    if not defaults or not path:
        return value
    container = defaults
    for key in path:
        if key not in container:
            return value
        container = container[key]
    return container


def argparse_options(
    parser, schema, defaults=None, attributes=None, default_from_schema=True
):

    """
    Add cli options to an argparse ArgumentParser instance

    **Arguments**

    - parser (`argparse.ArgumentParser`)
    - schema (`Schema`)

    **Keyword Arguments**

    - defaults (`dict`): if specified will override defaults from here
    - attributes (`list<str>`): can hold a list of attribute names.
    if specified only matching attributes will be aded
    - default_from_schema (`bool`): determines if defaults for
    argparser should come from the schema
    """

    def optionize(attribute, path):
        if not attribute.cli:
            return

        if attributes and destination_name(path) not in attributes:
            return

        kwargs = {
            "type": lambda x: attribute.validate(x, path),
            "help": attribute.help,
            "dest": destination_name(path),
        }

        if default_from_schema:
            kwargs["default"] = default(attribute.default, path, defaults)
        else:
            kwargs["default"] = default(None, path, defaults)

        name = option_name(path, delimiter=".")

        finalize = getattr(attribute, "finalize_argparse", None)
        if finalize:
            name = finalize(kwargs, name)

        if kwargs["default"] is not None and attribute.cli_show_default:
            kwargs["help"] = "{} ({})".format(kwargs["help"], kwargs["default"])

        parser.add_argument(name, **kwargs)

    schema.walk(optionize)


def apply_argparse(args, config):

    """
    Takes the output of a parser and applies it to a Config object.

    **Arguments**

    - args (`argparse.Namespace`):  the result of parser.parse_args()
    - config (`Config`): the config object

    **Returns**

    - config (`Config`): now updated with args
    """

    for k in args.__dict__:
        apply_arg(k, args, config)
    return config


def apply_arg(original_key, args, config):

    """
    Function for applying arguments to a config. Applies to nested
    configs as well.

    **Arguments**

    - original_key (`str`): the name of the argument in the argparse Namespace
    - args (`argparse.Namespace`): the entire Namespace, ie the result of the parser
    - config (`Config`): the config object

    """
    from confu.schema import Attribute, Schema

    schema = config._schema

    path = original_key.split("__")

    arg_data = getattr(args, original_key)

    if len(path) > 1:
        data = config.data

        current_schema = schema._attr

        for key in path:

            if isinstance(current_schema.get(key), Schema):
                current_schema = current_schema.get(key)._attr

            elif isinstance(current_schema.get(key), Attribute):
                attribute = current_schema.get(key)

                # If we cannot find the attribute in the schema
                # we don't add it
                if attribute is None:
                    return

        for key in path[:-1]:
            if key not in data:
                data[key] = {}
            data = data[key]
        data[path[-1]] = arg_data

    else:
        attribute = schema._attr.get(original_key)

        # If we cannot find the attribute in the schema
        # we don't add it
        if attribute is None:
            return

        config.data[original_key] = arg_data
        # attribute.validate(arg_data, path)


class click_options:

    """
    Add cli options to a click decorated function

    Use like a decorator

    **Arguments**

    - schema (`Schema`)

    **Keyword Arguments**

    - defaults (`dict`): if specified will override defaults from here
    - attributes (`list<str>`): can hold a list of attribute names.
    if specified only matching attributes will be aded
    """

    def __init__(self, schema, defaults=None, attributes=None):
        self.schema = schema
        self.defaults = defaults
        self.attributes = attributes

    def __call__(self, fn):
        container = {"fn": fn}
        defaults = self.defaults
        attributes = self.attributes

        def optionize(attribute, path):
            if not attribute.cli:
                return

            if attributes and destination_name(path) not in attributes:
                return

            def validate_and_convert(value):
                return attribute.validate(value, path)

            validate_and_convert.__name__ = attribute.__class__.__name__

            kwargs = {
                "type": click.types.FuncParamType(validate_and_convert),
                "help": attribute.help,
                "default": default(attribute.default, path, defaults),
            }

            name = option_name(path)

            finalize = getattr(attribute, "finalize_click", None)
            if finalize:
                name = finalize(kwargs, name)

            if kwargs["default"] is not None and attribute.cli_show_default:
                kwargs["help"] = "{} ({})".format(kwargs["help"], kwargs["default"])

            container["fn"] = click.option(
                name, destination_name(path), **kwargs
            ).__call__(fn)

        self.schema.walk(optionize)
        return container["fn"]



"""
utility functions / classes
"""
import os


def config_parser_dict(config):
    """
    Takes a configparser.ConfigParser instance and returns a dict
    of sections and their keys and values

    **Arguments**

    - config (`configparser.ConfigParsers`)

    **Returns**

    dict
    """
    return {s: dict(config.items(s)) for s in config.sections()}


_DEFAULT_ARG = object()


class SettingsManager:

    """
    Scoped settings management with environment variable override support
    """

    def __init__(self, scope):

        """
        Arguments:

        - scope (`dict`)
        """

        self.scope = scope

    def set_from_env(self, name, default=_DEFAULT_ARG):
        """
        Sets a scope variable from a environment variable of the same name.

        This is useful to leave the option unset and use default if it
        already exists in the scope (which may change).
        """
        if default is _DEFAULT_ARG and name not in os.environ:
            return

        self.scope[name] = os.environ.get(name, default)

    def set_option(self, name, value, envvar_type=None):
        """
        Sets an option, first checking for env vars,
        then checking for value already set,
        then going to the default value if passed.
        Environment variables are always strings, but
        we try to coerce them to the correct type first by checking
        the type of the default value provided. If the default
        value is None, then we check the optional envvar_type arg.
        """

        # If value is in True or False we
        # call set_bool to take advantage of
        # its type checking for environment variables
        if isinstance(value, bool):
            return self.set_bool(name, value)

        if value is not None:
            envvar_type = type(value)
        else:
            # If value is None, we'll use the provided envvar_type, if it is not None
            if envvar_type is None:
                raise ValueError(
                    f"If no default value is provided for the setting {name} the envvar_type argument must be set."
                )

        if name in os.environ:
            env_var = os.environ.get(name)
            # Coerce type based on provided value
            self.scope[name] = envvar_type(env_var)
        # If the environment variable isn't set
        else:
            self.set_default(name, value)

    def set_bool(self, name, value):
        """ Sets and option, first checking for env vars, then checking for value already set, then going to the default value if passed. """
        if name in os.environ:
            envval = os.environ.get(name).lower()
            if envval in ["1", "true", "y", "yes"]:
                self.scope[name] = True
            elif envval in ["0", "false", "n", "no"]:
                self.scope[name] = False
            else:
                raise ValueError(
                    "{} is a boolean, cannot match '{}'".format(name, os.environ[name])
                )
        self.set_default(name, value)

    def set_default(self, name, value):
        """ Sets the default value for the option if it's not already set. """
        if name not in self.scope:
            self.scope[name] = value



"""
Fundamental schema attributes

These can be imported directly from `confu.schema`
"""

import collections
import configparser
import copy
import os
from inspect import isclass

from confu.exceptions import ApplyDefaultError, ValidationError, ValidationWarning
from confu.util import config_parser_dict


class Attribute:

    """
    Base confu schema attribute. All other schema attributes
    should extend this

    """

    def __init__(self, name="", **kwargs):

        """
        Initialize attribute

        **Keyword Arguments**

        - name (`str`): describes the attribute name, if not specified
          explicitly will be set through the schema that instantiates
          the attribute.
        - default (`mixed`): the default value of this attribute. Once a default
          value is set, schema validation will no longer raise a
          validation error if the attribute is missing from the
          configuration.
        - choices (`list`): if specified on values in this list may be set
          for this attribute
        - help (`str`): help description
        - cli (`bool=True|function`): enable CLI support for this attribute
        - deprecated (`str`): version id of when this attribute will be deprecated
        - added (`str`): version id of when this attribute was added to the schema
        - removed (`str`): version id of when this attribute will be removed
        """

        # attribute name in the schema
        self.name = name

        # the default value to use, if specified a value is
        # not required in the config, other wise empty values
        # will raise a validation error during validation
        #
        # can be a function

        if "default" in kwargs:
            self.default_handler = kwargs.get("default")

        # deprecated in version
        self.deprecated = kwargs.get("deprecated")

        # added in version
        self.added = kwargs.get("added")

        # removed in version
        self.removed = kwargs.get("removed")

        # list of choices for value, if not empty values will
        # be validated against the choices in this list.
        self.choices_handler = kwargs.get("choices", [])

        # help text describing the attribute
        self.help = kwargs.get("help", "")

        # include this attribute when generating cli attributes
        self.cli_toggle = kwargs.get("cli", True)

        # show default value in cli help text
        self.cli_show_default = kwargs.get("cli_show_default", True)

        self.container = None

    @property
    def has_default(self):
        return hasattr(self, "default_handler")

    @property
    def default_is_none(self):
        return self.has_default and self.default is None

    @property
    def default(self):
        """
        Return the default value for this attribute
        """
        default = getattr(self, "default_handler", None)
        if callable(default):
            return default(self)
        return default

    @property
    def choices(self):
        """
        Return a list of possible value choices for this attribute

        Will return an empty list if the attribute is **NOT** limited by
        choices
        """

        choices_handler = getattr(self, "choices_handler", None)

        if callable(choices_handler):
            return self.choices_handler(self)
        return choices_handler

    @property
    def cli(self):
        """
        Returns whether or not the attribute is available
        in as a cli argument  when confu is used to generated
        argparse or click arguments
        """

        toggle = getattr(self, "cli_toggle", True)
        if callable(toggle):
            return toggle(self)
        return toggle

    def validate(self, value, path, **kwargs):
        """
        Validate a value for this attribute

        Will raise a `ValidationError` or `ValidationWarning` exception on
        validation failure

        **Arguments**

        - value (`mixed`): the value to validate
        - path (`list`): current path in the config schema, this is mostly
          used to identify where an error occured when validating
          against config data, you can pass an empty list to it when
          calling this manually
        """

        if not self.container and not self.name:
            raise ValidationError(
                self, path, value, "attribute at top level defined without a name"
            )

        if self.choices_handler:
            if value not in self.choices:
                raise ValidationError(self, path, value, "invalid choice")
        return value


class Str(Attribute):

    """
    Attribute that requires a string value

    Keyword Attributes:
        - blank <bool=False>: if True allow "" as a value
    """

    def __init__(self, name="", **kwargs):
        super().__init__(name=name, **kwargs)
        self.blank = kwargs.get("blank", False)
        if not self.blank and self.default_is_blank:
            self.blank = True

    @property
    def default_is_blank(self):
        return self.default == ""

    def validate(self, value, path, **kwargs):
        if not isinstance(value, str) and not self.default_is_none:
            raise ValidationError(self, path, value, "string expected")

        if value == "" and not self.blank:
            raise ValidationError(self, path, value, "cannot be blank")

        return super().validate(value, path, **kwargs)


class File(Str):
    """
    Attribute that requires a file to exist at path

    **Keyword Attributes**

    - require_exist ('bool=True'): if `True` file needs to exist
    """

    def __init__(self, name="", **kwargs):
        super().__init__(name=name, **kwargs)
        self.require_exist = kwargs.get("require_exist", True)

    def validate(self, value, path, **kwargs):
        value = super().validate(value, path, **kwargs)

        if value is None and self.default_is_none:
            return value

        if value == "" and self.blank:
            return value

        # make sure env vars get expanded
        value = os.path.expandvars(value)

        # make sure user vars get expanded
        value = os.path.expanduser(value)

        # expand to absolute path
        value = os.path.abspath(value)

        valid = (os.path.exists(value) or not self.require_exist) and not os.path.isdir(
            value
        )
        if not valid:
            raise ValidationError(self, path, value, "file does not exist")

        return value


class Directory(Str):

    """
    Attribute that requires an existing directory path

    **Keyword Arguments**

    - require_exist ('bool=True'): if `True` directory needs to exist
    - create (`octal`): if set, instead of raising a ValidationError
    on a non-existing directory, attempt to create directory first
    using the value passed as mode (chmod) e.g., create=0o777
    """

    def __init__(self, name="", **kwargs):
        super().__init__(name=name, **kwargs)

        self.create = kwargs.get("create")
        self.require_exist = kwargs.get("require_exist", True)

    def makedir(self, value, config_path):
        try:
            os.makedirs(value, self.create)
        except Exception as err:
            raise ValidationError(
                self,
                config_path,
                value,
                "tried to create directory  but failed with error" ": {}".format(err),
            )

    def validate(self, value, path, **kwargs):
        value = super().validate(value, path, **kwargs)

        if value is None and self.default_is_none:
            return value

        # make sure env vars get expanded
        value = os.path.expandvars(value)

        # if value is blank, and validation was not caught by `blank` validation
        # of Str we forfeit validation and it is assume that the application
        # will validate manually once the path is set.
        if value == "":
            return value

        # make sure user vars get expanded
        value = os.path.expanduser(value)

        # expand to absolute path
        value = os.path.abspath(value)

        if self.create is not None and not os.path.exists(value):
            self.makedir(value, path)

        if self.require_exist:
            valid = os.path.exists(value) and os.path.isdir(value)
        else:
            valid = True

        if not valid:
            raise ValidationError(
                self, path, value, f"valid path to directory expected: {value}"
            )

        return value


class Bool(Attribute):
    """
    Attribute that requires a boolean value
    """

    true_values = ["true", "yes", "1"]
    false_values = ["false", "no", "0"]

    def __init__(self, name="", **kwargs):
        super().__init__(name=name, **kwargs)
        self.cli_show_default = False

    def validate(self, value, path, **kwargs):
        if isinstance(value, str):
            if value.lower() in self.true_values:
                value = True
            elif value.lower() in self.false_values:
                value = False
            else:
                raise ValidationError(self, path, value, "boolean expected")
        return super().validate(bool(value), path, **kwargs)

    def finalize_click(self, param, name):
        del param["type"]
        return "{}/--no-{}".format(name, name.strip("-"))

    def finalize_argparse(self, param, name):
        del param["type"]
        if not param["default"]:
            param.update(action="store_true")
        else:
            param.update(action="store_false")
            if param.get("help"):
                param["help"] = "DISABLE {}".format(param["help"])
            name = "--no-{}".format(name.strip("-"))
        return name


class Int(Attribute):

    """
    Attribute that requires an integer value
    """

    def validate(self, value, path, **kwargs):
        if value is None and self.default_is_none:
            return value
        try:
            value = int(value)
        except (TypeError, ValueError):
            raise ValidationError(self, path, value, "integer expected")
        return super().validate(value, path, **kwargs)


class Float(Attribute):

    """
    Attribute that requires a float value
    """

    def validate(self, value, path, **kwargs):
        if value is None and self.default_is_none:
            return value
        try:
            value = float(value)
        except (TypeError, ValueError):
            raise ValidationError(self, path, value, "float expected")
        return super().validate(value, path, **kwargs)


class List(Attribute):

    """
    Attribute that requires a list value
    """

    # TODO: item should be a required positional argument, but
    # doing so means flipping the order with name, which breaks
    # peoples schemas (major version fix?)

    def __init__(self, name=None, item=None, **kwargs):
        """
        Initialize List attribute

        **Keyword Arguments**

        - name (`str`): describes the attribute name, if not specified
          explicitly will be set through the schema that instantiates
          the attribute.
        - item (`Attribute`): allows you to specify an arbitrary attribute
          to use for all values in the list.
        - default (`mixed`): the default value of this attribute. Once a default
          value is set, schema validation will no longer raise a
          validation error if the attribute is missing from the
          configuration.
        - help (`str`): help description
        - cli (`bool=True`): enable CLI support for this attribute
        - deprecated (`str`): version id of when this attribute will be deprecated
        - added (`str`): version id of when this attribute was added to the schema
        - removed (`str`): version id of when this attribute will be removed
        """

        if not isinstance(item, Attribute):
            raise TypeError("item needs to be a confu attribute")

        if "default" not in kwargs:
            kwargs["default"] = []

        if isclass(item):
            kwargs["cli"] = False

        if isinstance(item, Attribute):
            item.container = self

        super().__init__(name, **kwargs)

        self.item = item

    @property
    def cli(self):
        if isinstance(self.item, Schema):
            return False
        return super().cli

    def validate(self, value, path, **kwargs):

        if isinstance(value, str):
            value = value.split(",")

        if not isinstance(value, list):
            raise ValidationError(self, path, value, "list expected")

        errors = kwargs.get("errors", ValidationErrorProcessor())
        warnings = kwargs.get("warnings", ValidationErrorProcessor())

        validated = []
        idx = 0
        for item in value:
            try:
                if isinstance(self.item, Schema):
                    validated.append(
                        self.item.validate(
                            item, path + [idx], errors=errors, warnings=warnings
                        )
                    )
                else:
                    validated.append(self.item.validate(item, path + [idx]))
                idx += 1
            except ValidationError as error:
                errors.error(error)
            except ValidationWarning as warning:
                warnings.warning(warning)
        return super().validate(validated, path, **kwargs)


class ValidationErrorProcessor:
    """
    This the default validation error processor, it will raise an exception
    when a warning or error is encountered
    """

    def error(self, error):
        raise error

    def warning(self, warning):
        raise warning


class CollectValidationExceptions(ValidationErrorProcessor):
    """
    This validation error processor will store all errors and warnings it encounters
    and NOT raise any exceptions
    """

    def __init__(self):
        self.exceptions = []

    def __iter__(self):
        yield from self.exceptions

    def __len__(self):
        return len(self.exceptions)

    def __getitem__(self, key):
        return self.exceptions[key]

    def error(self, error):
        self.exceptions.append(error)

    def warning(self, warning):
        self.exceptions.append(warning)


class Schema(Attribute):

    """
    Describes a confu schema.

    Instantiate confu attributes as properties of the schema.

    As the schema itself is a confu attribute, you may nest schemas
    within schemas

    **Example**

    ```
    class MySchema(Schema):
        str_attr = Str()
    ```
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize schema

        **Keyword Arguments**

        - item (`Attribute`): allows you to specify an arbitrary attribute
          to use for all values in the schema. This is only allowed if your
          schema does **NOT** explicitly set any attributes in it's definition
        - name (`str`): describes the attribute name, if not specified
          explicitly will be set through the schema that instantiates
          the attribute.
        - default (`mixed`): the default value of this attribute. Once a default
          value is set, schema validation will no longer raise a
          validation error if the attribute is missing from the
          configuration.
        - help (`str`): help description
        - cli (`bool=True`): enable CLI support for this attribute
        - deprecated (`str`): version id of when this attribute will be deprecated
        - added (`str`): version id of when this attribute was added to the schema
        - removed (`str`): version id of when this attribute will be removed
        """

        # collect attributes
        self._attr = {}
        for name in dir(self):
            attr = getattr(self, name)
            if isinstance(attr, Attribute):
                self._attr[name] = attr
                if not attr.name:
                    attr.name = name

        if "default" not in kwargs:
            kwargs["default"] = {}

        kwargs["cli"] = False

        self.item = kwargs.get("item")
        if self.item and self._attr:
            raise ValueError(
                "You cannot specify an `item` attribute on a "
                "Schema instance that has attributes defined within"
            )
        elif self.item:
            self.item.container = self

        super().__init__(*args, **kwargs)

    def attributes(self):
        # redundant?
        yield from list(self._attr.items())

    def walk(self, callback, path=None):
        if not path:
            path = []
        for name, attribute in self.attributes():
            if isinstance(attribute, Schema):
                callback(attribute, path + [name])
                attribute.walk(callback, path=path + [name])
            else:
                callback(attribute, path + [name])

    def validate(self, config, path=None, errors=None, warnings=None):

        """
        Validate config data against this schema

        **Attributes**

        - config (`dict`): config to validate

        **Keyword Attributes**

        - path (`list`): current path in the config data, this can be
          ignored on the initial call and will be set automatically
          on any subsequent calls (nested schemas)
        - errors (`ValidationErrorProcessor`)
        - warnigns (`ValidationErrorProcessor`)
        """

        if path is None:
            path = []
        if errors is None:
            errors = ValidationErrorProcessor()
        if warnings is None:
            warnings = ValidationErrorProcessor()

        # munge Config support without having to import munge
        if isinstance(config, collections.MutableMapping) and hasattr(config, "data"):
            config = config.data
        elif isinstance(config, configparser.ConfigParser):
            config = config_parser_dict(config)

        if not isinstance(config, dict):
            return errors.error(
                ValidationError(path[-1], path, config, "dictionary expected")
            )

        for key, value in list(config.items()):
            try:
                attribute = self._attr.get(key, self.item)

                if attribute is None:
                    raise ValidationWarning(
                        key, path, value, f"unknown attribute '{key}'"
                    )
                else:
                    config[key] = attribute.validate(
                        value, path + [key], errors=errors, warnings=warnings
                    )
            except ValidationError as error:
                errors.error(error)
            except ValidationWarning as warning:
                warnings.warning(warning)

        for name, attribute in self.attributes():
            if name not in config and not attribute.has_default:
                errors.error(ValidationError(attribute, path + [name], None, "missing"))

        return config


class Dict(Schema):
    """
    Wrapper for schema with arbitrary keys

    For this the `item` property needs to be set.
    """

    def __init__(self, name=None, item=None, *args, **kwargs):
        super().__init__(name=name, item=item, *args, **kwargs)


class ProxySchema(Schema):

    """
    An object that can be used in place of a schema in order to
    dynamically obtain a schema instance from somewhere else during
    validate
    """

    def schema(self, config):
        """
        return a schema instance
        """
        raise NotImplementedError()

    def validate(self, config, path=None, errors=None, warnings=None):
        """
        call validate on the schema returned by self.schema
        """
        return self.schema(config).validate(
            config, path=path, errors=errors, warnings=warnings
        )


def validate(schema, config, raise_errors=False, log=None, **kwargs):
    """
    Helper function that allows schema validation to either collect or raise errors

    **Arguments**

    - schema (`Schema`): schema instance
    - config (`dict|munge.Config`)

    **Keyword Arguments**

    - raise_errors (`bool=False`): if `True` will raise a ValidationError exception
      if it encounters validation errors

      If `False` it will instead collect errors and warnings and return a tuple:

      ```
      success(bool), errors(CollectValidationExceptions), warnings(CollectValidationException)
      ```

    - log (`callable`): function to use to log errors, will be passed
      a str message
    - any additional kwargs will be passed on to `Schema.validate`
    """

    warnings = CollectValidationExceptions()
    if raise_errors:
        schema.validate(config, warnings=warnings, **kwargs)
        return (True, [], warnings)
    else:
        errors = CollectValidationExceptions()
        schema.validate(config, errors=errors, warnings=warnings, **kwargs)

        num_errors = len(errors)
        num_warnings = len(warnings)

        success = num_errors == 0

        if log and callable(log):
            for error in errors:
                log(f"[Config Error] {error.pretty}")
            for warning in warnings:
                log(f"[Config Warning] {warning.pretty}")
            if not success:
                log(f"{num_errors} errors, {num_warnings} warnings in config")

        return (success, errors, warnings)


def apply_default(config, attribute, path):
    """
    Apply attribute default to config dict at the specified path

    **Arguments**

    - config (`dict`): the config dictonary
    - attribute (`Attribute`): attribute instance
    - path (`list(str)`): full path of the attribute in the schema
    """

    _config = config
    prev = None

    for section in path:
        prev = _config
        _config = _config.get(section)

    if isinstance(attribute, List):

        # attribute is a List, need to handle items
        # accordingly
        if _config and isinstance(attribute.item, Schema):

            # list is holding schemas, apply defaults
            # to each item in the list
            for item in _config:
                apply_defaults(attribute.item, item, debug=True)

        if _config and isinstance(attribute.item, List):

            # list is holding lists, apply defaults
            # to each item in the list
            for item in _config:
                apply_default(item, attribute.item, [])

        elif _config is None and attribute.has_default:

            # list is holding normal attribute, set default
            # value
            prev[section] = attribute.default

    elif isinstance(attribute, Schema):

        # attribute is a Schema

        if isinstance(attribute, ProxySchema):
            attribute = attribute.schema(_config)

        if (
            _config
            and isinstance(attribute.item, List)
            and isinstance(attribute.item.item, Schema)
        ):
            # schema with arbitrary keys
            # holding lists holding schemas
            # TODO: find a cleaner way to handle this case

            for k, item in list(_config.items()):
                apply_default(_config, attribute.item, [k])

        if _config is None:
            prev[section] = copy.deepcopy(attribute.default or {})

        if attribute.item is None:
            apply_defaults(attribute, prev[section])
        if isinstance(attribute.item, Schema):
            apply_defaults(attribute.item, prev[section])

    elif _config is None and attribute.has_default:
        prev[section] = attribute.default


def apply_defaults(schema, config, debug=False):
    """
    Take a config object and apply a schema's default values to keys that
    are missing.

    **Arguments**

    - schema (`Schema`): schema instance
    - config (`dict`): the config dictonary
    """

    if isinstance(schema, ProxySchema):
        # schema is proxy schema, retrieve actual schema
        # before proceeding
        schema = schema.schema(config)

    if isinstance(schema.item, Schema):
        # schema has arbitrary keys holding another schema
        for k, v in list(config.items()):
            apply_defaults(schema.item, v, debug=debug)
        return
    elif isinstance(schema.item, List):
        # schema has arbitrary keys holding a list
        for k, v in list(config.items()):
            apply_default(config, schema.item, [k])
        return

    # normal schema, walk it's attributes and apply defaults
    def callback(attribute, path):
        try:
            apply_default(config, attribute, path)
        except Exception as exc:
            raise ApplyDefaultError(attribute, path, None, exc)

    schema.walk(callback)



"""
Schema to data generators
"""
from confu.schema import Attribute, Schema


class ConfigGenerator:
    """
    Generate config from schema using default values
    """

    def __init__(self):
        pass

    def generate(self, schema):

        """
        Generate confug from schema using default values

        **Arguments**

        - schema (`Schema|Attribute`): confu schema object

        **Returns**

        generated config `dict`
        """

        if isinstance(schema, Schema):
            config = {}
            for name, attribute in schema.attributes():
                config[name] = self.generate(attribute)
            return config
        elif isinstance(schema, Attribute):
            return self.generate(schema.default)
        elif isinstance(schema, list):
            return [self.generate(item) for item in schema]
        return schema


def generate(schema, generator=None):
    """
    generate config shortcut function

    **Arguments**

    - schema (`Schema`): confu schema object

    **Keyword Arguments**

    - generator (`ConfigGenerator`): generator object, if non is supplied
    will instantiate a ConfigGenerator instance itself

    **Returns**

    generated config (`dict`)
    """
    if not generator:
        generator = ConfigGenerator()
    return generator.generate(schema)



"""
config management
"""


import collections
import copy

import confu.schema


class Config(collections.Mapping):
    """
    class for storing and manipulating config data
    """

    def __init__(self, schema, data=None, meta=None):
        """
        **Arguments**

        - schema (`confu.schema`): schema object

        **Keyword Arguments**

        - data (`dict`): dict to set initial data
        - meta (`dict`): any additional metadata to pass along with config
        """
        self._base_data = None
        self._data = None
        self._schema = schema
        self.meta = meta if meta else {}

        self.errors = []
        self.warnings = []
        self.valid = None

        # use property setter to apply
        self.data = data if data else {}

    def copy(self):
        """ return a read only copy of data """
        return copy.deepcopy(self.data)

    @property
    def data(self):
        """ config data, should be used for read only """
        if self._data:
            return self._data

        data = copy.deepcopy(self._base_data)

        try:
            confu.schema.apply_defaults(self._schema, data)
        except confu.exceptions.ApplyDefaultError as exc:
            self.apply_default_error = exc

        self.valid, self.errors, self.warnings = confu.schema.validate(
            self.schema, data
        )

        if hasattr(self, "apply_default_error"):
            self.errors.exceptions.insert(0, self.apply_default_error)

        self._data = data
        return self._data

    @data.setter
    def data(self, value):
        self._base_data = value
        self._data = None

    @property
    def schema(self):
        """ return a read only copy of schema """
        return copy.deepcopy(self._schema)

    def get_nested(self, *args):
        """
        get a nested value, returns None if path does not exist
        """
        data = self.data
        for key in args:
            if key not in data:
                return None
            data = data[key]
        return data

    def __getitem__(self, key):
        return self.data[key]

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)



class SoftDependencyError(ImportError):
    """
    Raised when a feature requires a dependency that is missing
    """

    def __init__(self, dep_name):
        super().__init__(f"To use this feature this dependency is required: {dep_name}")


class ValidationErrorBase(ValueError):
    """
    Config validation error interface
    """

    def __init__(self, attribute, path, value, reason):
        """
        **Arguments**

        - attribute (`Attribute`): confu attribute instance
        - path (`list`): attribute path
        - value (`mixed`): value that caused the validation error
        - reason (`str`): human readable reason message for validation error
        """

        msg = f"{path}: {reason}"
        self.details = {
            "path": path,
            "attribute": attribute,
            "value": value,
            "reason": reason,
        }
        super().__init__(msg)

    @property
    def pretty(self):
        """
        pretty formatted error message
        """
        return "{}: {}".format(
            ".".join([str(i) for i in self.details["path"]]), self.details["reason"]
        )

    def __eq__(self, other):
        if type(other) != type(self):
            return False

        return (
            self.details["path"] == other.details["path"]
            and self.details["value"] == other.details["value"]
            and self.details["reason"] == other.details["reason"]
        )


class ValidationWarning(ValidationErrorBase):
    """
    Config validation warning
    """

    pass


class ValidationError(ValidationErrorBase):
    """
    Config validation error
    """

    pass


class ApplyDefaultError(ValidationErrorBase):
    """
    Raised when an exception occured during apply_defaults
    """

    def __str__(self):
        return self.pretty



"""
Attributes that deal with networking specific values such as emails, urls and ip addresses

These can be imported directly from `confu.schema`

## Requirements

- `ipaddress` for ip address validation
"""

import ipaddress
import re
from urllib.parse import urlparse

from confu.exceptions import SoftDependencyError
from confu.schema.core import Str, ValidationError


class Email(Str):

    """
    Describes an email address
    """

    def validate(self, value, path, **kwargs):
        value = super().validate(value, path, **kwargs)

        if value == "" and self.blank:
            return value

        if value is None and self.default_is_none:
            return value

        # TODO: any reason to get more sophisticated than this?
        if not re.match(r"[^@\s]+@[^@\s]+", value):
            raise ValidationError(self, path, value, "email address expected")

        return value


class Url(Str):

    """
    Describes a URL
    """

    def __init__(self, name="", **kwargs):
        super().__init__(name=name, **kwargs)
        self.schemes = kwargs.get("schemes", [])

    def validate(self, value, path, **kwargs):
        """
        Currently only validates by running urlparse against it
        and checking that a scheme and netloc is set - and if a list of allowed
        schemes is provide that the scheme is valid against that list

        TODO: may want something more sophisticated than that - could look
        at django's url validator
        """
        value = super().validate(value, path, **kwargs)

        if value == "" and self.blank:
            return value

        if value is None and self.default_is_none:
            return value

        try:
            result = urlparse(value)
        except ValueError:
            raise ValidationError(self, path, value, "url expected")

        if not result.scheme:
            raise ValidationError(self, path, value, "no url scheme specified")

        if not result.netloc:
            raise ValidationError(self, path, value, "no url netloc specified")

        if self.schemes and result.scheme not in self.schemes:
            raise ValidationError(
                self, path, value, f"invalid url scheme: {result.scheme}"
            )

        return value


class IpAddress(Str):

    """
    Describes a IPv4 or IPv6 address
    """

    def __init__(self, name="", protocol=None, **kwargs):

        """
        Initialize attribute

        **Keyword Arguments**

        - name (`str`): describes the attribute name, if not specified
          explicitly will be set through the schema that instantiates
          the attribute.
        - protocol (`int`): ip version, can be 4, 6 or None - if it is none
          the attribute can hold either a v4 or a v6 IP address.
        - default (`mixed`): the default value of this attribute. Once a default
          value is set, schema validation will no longer raise a
          validation error if the attribute is missing from the
          configuration.
        - choices (`list`): if specified on values in this list may be set
          for this attribute
        - help (`str`): help description
        - cli (`bool=True`): enable CLI support for this attribute
        - deprecated (`str`): version id of when this attribute will be deprecated
        - added (`str`): version id of when this attribute was added to the schema
        - removed (`str`): version id of when this attribute will be removed
        """

        super().__init__(name=name, **kwargs)
        if not ipaddress:
            raise SoftDependencyError("ipaddress")
        if protocol not in [None, 4, 6]:
            raise ValueError("IpAddress protocol needs to be either 4, 6 or None")
        self.protocol = protocol

    def validate_v4(self, value, path, **kwargs):
        try:
            return ipaddress.IPv4Address(value)
        except ipaddress.AddressValueError:
            return False

    def validate_v6(self, value, path, **kwargs):
        try:
            return ipaddress.IPv6Address(value)
        except ipaddress.AddressValueError:
            return False

    def validate(self, value, path, **kwargs):
        value = super().validate(value, path, **kwargs)

        if value is None and self.default_is_none:
            return value

        if self.blank and value == "":
            return value

        value = f"{value}"
        value_v4 = self.validate_v4(value, path, **kwargs)
        value_v6 = self.validate_v6(value, path, **kwargs)
        if self.protocol == 4 and not value_v4:
            raise ValidationError(self, path, value, "invalid ip (v4)")
        elif self.protocol == 6 and not value_v6:
            raise ValidationError(self, path, value, "invalid ip (v6)")
        elif self.protocol is None and not value_v4 and not value_v6:
            raise ValidationError(self, path, value, "invalid ip (v4 or v6)")
        return value_v4 or value_v6



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['confu', 'confu.schema']

package_data = \
{'': ['*']}

install_requires = \
['markdown-include>=0.5,<1']

extras_require = \
{':extra == "docs"': ['mkdocs>=1.0.0,<2.0.0', 'pymdgen<2']}

setup_kwargs = {
    'name': 'confu',
    'version': '1.7.0',
    'description': 'Configuration file validation and generation',
    'long_description': None,
    'author': '20C',
    'author_email': 'code@20c.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/20c/confu',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
